﻿
// Debian GNU/Linux on WSL
// gcc mp3_14_mkfl.c -lid3tag -lz
// Make file list

#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <iconv.h>
#include <id3tag.h>			// /usr/local/include
#include <time.h>

#include "mp3_14.h"

#define		SNDTAGINFO				"SNDTAG14.DAT"		// タグ情報出力先

#define		RANDOMSPECFILE			"randomspec.txt"	// 特定ランダム再生ディレクトリリストファイル(未使用)
#define		MAXRANDOMSPECDIRS		5					// 特定ランダム再生ディレクトリ最大数

#define		ENV_RANDOM				"MP3_14_RANDOM"


// ディレクトリ構成解析作業用
typedef struct _DirList {
	char		displayname[DL_DISPLAYNAME_LEN + 1];	// ディレクトリ表示名(スペース埋め)
	char		fulldir[DIRPATH_LEN + 1];				// ディレクトリフルパス
	uint16_t	lenfulldir;								// 同バイト数
	int			idx_dirname;			// fulldirのパス名を除いたDIR名の開始位置
	uint16_t	recno_subdir_start;		// サブDIR開始行
	uint16_t	num_subdir;				// サブDIR数
	uint16_t	recno_parentdir;		// 親DIR行
	uint16_t	recno_grandparentdir;	// 親親DIR行
	//uint16_t	recno_nextrandomplay;	// ランダム再生時、次のレコードNo (mp3_14では常に0)
	uint16_t	num_sndfile;			// MP3ファイル数
	uint16_t	recno_sndfile;			// MP3ファイルリスト先頭レコード番号
} DIRLIST;


// MP3ファイルリスト構造体
typedef struct _Mp3List {
	char	TitleName[ML_TITLENAME_LEN + 1];	// 表示名
	char	FileName[ML_FILENAME_LEN];
} MP3LIST14;


// ランダム再生順
typedef struct _RandomList {
	uint8_t		randomtype;		// 再生単位 0=フォルダ単位、1=トラック単位
	uint16_t	dirno;			// DirList[]のレコード番号(0-)
	uint16_t	parentdirno;	// 親ディレクトリ
	uint16_t	trackno;		// MP3ファイルリストのレコード番号
} RANDOMLIST;

#define			MAX_RANDOMLIST		(MAX_DIRS + MAXRANDOMSPECDIRS*50)		// 特定ランダム再生ディレクトリあたり平均50曲の想定

//
FILE			*Fml;
DIRLIST			DirList[MAX_DIRS];
RANDOMLIST		RandomList[MAX_RANDOMLIST];
uint16_t		NumRandomList;
uint16_t		DirListDetailRecNum;	// DIR構成ファイルの明細レコード総数
uint16_t		NumSndDir;
uint16_t		NumRandomDir;
uint8_t			NumIndexes;
char			CrDateTime[DL_CRDATETIME_LEN];

char			RandomSpecDirs[MAXRANDOMSPECDIRS][DL_FULLDIR_LEN];
int				RandomSpecDirsLen[MAXRANDOMSPECDIRS];
int				NumRandomSpecDirs;


iconv_t icdUCS4toUTF8;
iconv_t icdUTF8toSJIS;
iconv_t icdUTF16LEtoSJIS;

extern char **environ;		// 環境変数取得

void ReadRandomSpecFile();
int ReadDirectoryStructure();
int compare_DIRLIST(const void *a, const void *b);
int WriteDirStructureList();
uint16_t MakeSndFileList(char *inFolder);
int compare_MP3LIST14(const void *a, const void *b);
int iconvUCS4toUTF8(id3_ucs4_t *ucs4, char *utf8str, int bufsize);
int iconvUTF8toSJIS(char *utf8str, int convbytes, char *sjisstr, int sjismaxlen);
int iconvUTF16LEtoSJIS(char *ucs4str, int convbytes, char *sjisstr, int sjismaxlen);
int convUCS4toUTF16LE(id3_ucs4_t *ucs4, char *utf16lestr, int bufsize);
void convUCS4toASCII(id3_ucs4_t *ucs4, char *asciistr, int bufsize);
void getID3TagInfo(char *inFolder, char *mp3, char *title, int titlebufsize, int *titlebytes);
void getID3TagInfoTest(FILE *fout, char *inFolder, char *mp3);
int SetRandomOrder();
void WriteRandomOrderFile();
void CreateHistFile();
char *GetEnvRandomDir();


int main(int argc, char *argv[]){
	char *env;
	icdUCS4toUTF8 = iconv_open("UTF-8", "UCS-4LE");
	icdUTF8toSJIS = iconv_open("Shift_JIS//TRANSLIT", "UTF-8");
	//icdUTF16LEtoSJIS = iconv_open("Shift_JIS//TRANSLIT", "UTF-16");
	
	env = GetEnvRandomDir();
	if (env != NULL){
		char	*c1, *c2;
		int		i = 0;
		int		loop_continue = 1;
		int		len;
		
		// "特定ランダム再生ディレクトリ"を環境変数によって渡す。複数ディレクトリは;（セミコロン）で区切る。
		printf("env=%s\n", env);
		c1 = env;
		while (loop_continue){
			c2 = strchr(c1, ';');
			if (c2 == NULL){
				loop_continue = 0;
			} else {
				*c2 = '\0';
			}
			len = strlen(c1);
			if (len > 0){
				strcpy(RandomSpecDirs[i], c1);
				RandomSpecDirsLen[i] = len;
				printf("spec dir=%s\n", RandomSpecDirs[i]);
				i++;
				if (i == MAXRANDOMSPECDIRS) break;
			}
			c1 = c2 + 1;
		}
		NumRandomSpecDirs = i;
		
	}
	
#if 0
		// TEST
		strcpy(RandomSpecDirs[0], "/snd/その他音楽/LP/LetItBe");
		RandomSpecDirsLen[0] = strlen(RandomSpecDirs[0]);
		strcpy(RandomSpecDirs[1], "/snd/その他音楽/LP/Mantovani");
		RandomSpecDirsLen[1] = strlen(RandomSpecDirs[1]);
		NumRandomSpecDirs = 2;
#endif
	
	if (argc > 1){
		memcpy(CrDateTime, argv[1], DL_CRDATETIME_LEN);
	} else {
		memset(CrDateTime, 0, DL_CRDATETIME_LEN);
	}
	
	/*
	 * ディレクトリ構造を読み取る
	 */
	ReadDirectoryStructure();
	
	/*
	 *	ランダム再生の順序を決める
	 */
	SetRandomOrder();
	WriteRandomOrderFile();
	
	
	/*
	 *	ディレクトリ構成ファイルを書き出す
	 */
	WriteDirStructureList();
	
	
	// 履歴ファイル
	CreateHistFile();
	
	iconv_close(icdUCS4toUTF8);
	iconv_close(icdUTF8toSJIS);
	//iconv_close(icdUTF16LEtoSJIS);
}


/*
 *	特定ランダム再生ディレクトリリストファイル
 */
void ReadRandomSpecFile(){
	FILE	*fin;
	char	buff[DL_FULLDIR_LEN];
	char	*c;
	int		i;
	
	fin = fopen(RANDOMSPECFILE, "r");
	if (fin == NULL){
		printf("cannot open input file\n");
		return;
	}
	i = 0;
	while (fgets(buff, DL_FULLDIR_LEN, fin) != NULL){
		printf("%s", buff);
		// 改行コードを取る
		c = strchr(buff, '\n');		// 0x0A
		if (c != NULL){
			*c = '\0';
		}
		strcpy(RandomSpecDirs[i], buff);
		i++;
		if (i == MAXRANDOMSPECDIRS) break;
	}
	NumRandomSpecDirs = i;
	fclose(fin);
}


/*
 *	ディレクトリ構造を読み取り、内部構造体にセットする。
 *	MP3ファイルの存在するディレクトリについて「音源ファイル(MP3)リスト」を作成していく。
 */
int ReadDirectoryStructure(){
	FILE	*fin;
	char	buff[1024];
	uint16_t	rowin;
	uint16_t	idxout;
	char	*c;
	char	*fullp;
	char	*parentp;
	char	*name;
	int		i, p;
	int		len;
	uint16_t	recno_sndfile;
	
	printf("ReadDirectoryStructure\n");
#if 0
	// linuxでは find /snd -maxdepth 10 -type d
	
	// DIRコマンド /AD=ディレクトリに関して、/ON=名前順、/S=サブディレクトリも、/B=ファイル名
	// << 同一ディレクトリ下のサブディレクトリは連続することが以下のプログラムの前提になっている。>>
	_stprintf(buff, _T("dir /AD /ON /S /N /B > %s%s"), TempDir, DIRLISTFILE);
	_tsystem(buff);
#endif

	fin = fopen(DIRLISTFILE, "r");
	if (fin == NULL){
		printf("cannot open input file\n");
		return -1;
	}
	Fml = fopen(SNDFILELISTFILE, "w");
	if (Fml == NULL){
		printf("cannot open mp3-list file\n");
		return -1;
	}

	NumSndDir = 0;		// MP3ファイルを持つDIR数
	NumRandomList = 0;	// ランダム再生単位数
	rowin = 0;
	idxout = 0;
	recno_sndfile = 0;

	while (fgets(buff, 1024, fin) != NULL){
		printf("%s", buff);
		// 改行コードを取る
		c = strchr(buff, '\n');		// 0x0A
		if (c != NULL){
			*c = '\0';
		}
		fullp = buff;
		if (strlen(fullp) > DIRPATH_LEN) continue;

		// fullpはディレクトリのフルパスである。
		// これをパス無しサブディレクトリ名と親ディレクトリパスとに分離する
		c = strrchr(fullp, '/');
		if (c != NULL){
			strcpy(DirList[idxout].fulldir, fullp);
			DirList[idxout].lenfulldir = strlen(DirList[idxout].fulldir);
			*c = '\0';
			name = c + 1;
			DirList[idxout].idx_dirname = (int)(name - fullp);
			if (name[0] == '_') continue;		// _ で始まるディレクトリは無視
			
			memset(DirList[idxout].displayname, ' ', DL_DISPLAYNAME_LEN);
			//		printf("iconvUTF8toSJIS start\n");
			iconvUTF8toSJIS(name, strlen(name), DirList[idxout].displayname, DL_DISPLAYNAME_LEN);
			//		printf("iconvUTF8toSJIS end\n");
			DirList[idxout].displayname[DL_DISPLAYNAME_LEN] = '\0';
			DirList[idxout].recno_subdir_start = 0;
			DirList[idxout].num_subdir = 0;
			DirList[idxout].recno_parentdir = 0;
			DirList[idxout].recno_grandparentdir = 0;
			DirList[idxout].num_sndfile = 0;
			DirList[idxout].recno_sndfile = 0;
			parentp = fullp;

			// DirListから親DIRを探す
			for (i=0; i<idxout; i++){
				//printf("[%s][%s]\n", parentp, DirList[i].fulldir);
				if (strcmp(parentp, DirList[i].fulldir) == 0){
					DirList[idxout].recno_parentdir = i + 1;		// 親DIRの位置
					// +1しているのはqsort実行時にルートディレクトリとルートディレクトリ直下のディレクトリを区別するため。
					// ここでrecno_parentdirに設定する値は暫定値。後にDirListをソートしてから親DIRを再設定する
					break;
				}
			}
			if (i == idxout){
				// 親が見つからないのはidxout==0すなわちルートDIR
			}
			
			DirList[idxout].num_sndfile = MakeSndFileList(DirList[idxout].fulldir);		// 音源ファイルリストを生成
			if (DirList[idxout].num_sndfile > 0){
				DirList[idxout].recno_sndfile = recno_sndfile;
				recno_sndfile += DirList[idxout].num_sndfile;
			}
			rowin++;
			idxout++;
		}
	}
	fclose(fin);
	fclose(Fml);
	
	DirListDetailRecNum = idxout;
	NumSndDir = 0;
	NumRandomList = 0;
	
	// 共通の親DIRを持つディレクトリが連続するように並べ替える。
	qsort(DirList, DirListDetailRecNum, sizeof(DIRLIST), compare_DIRLIST);
	// 親DIRの再検索
	for (i=0; i<DirListDetailRecNum; i++){
		c = strrchr(DirList[i].fulldir, '/');
		if (c != NULL){
			// 親DIR
			memcpy(buff, DirList[i].fulldir, (int)(c - DirList[i].fulldir));
			buff[(int)(c - DirList[i].fulldir)] = '\0';
			// 親DIRを探す
			for (p=0; p<i; p++){
				if (strcmp(buff, DirList[p].fulldir) == 0){
					if (DirList[p].num_subdir == 0){
						DirList[p].recno_subdir_start = i;	// 親DIRの最初のサブDIR
					}
					DirList[p].num_subdir++;			// 親DIRのサブDIR数
					DirList[i].recno_parentdir = p;		// 親DIRの位置
					DirList[i].recno_grandparentdir = DirList[p].recno_parentdir;	// 親親DIRの位置
					break;
				}
			}
			
			// 
			if (DirList[i].num_sndfile > 0 && NumRandomList < MAX_RANDOMLIST){
				// ランダム再生順作成データ
				int		j;
				int		rnd_flg = 0;
				for (j=0; j<NumRandomSpecDirs; j++){
					if (strncmp(DirList[i].fulldir, RandomSpecDirs[j], RandomSpecDirsLen[j]) == 0){
						rnd_flg = 1;
						break;
					}
				}
				if (rnd_flg == 1){		// 特定ランダム再生ディレクトリ:トラック単位でランダム再生する
					int	trno;
					for (trno=0; trno<DirList[i].num_sndfile; trno++){
						RandomList[NumRandomList].randomtype = 1;
						RandomList[NumRandomList].dirno = i;
						RandomList[NumRandomList].parentdirno = DirList[i].recno_parentdir;
						RandomList[NumRandomList].trackno = trno;
						NumRandomList++;
						if (NumRandomList == MAX_RANDOMLIST) break;
					}
				} else {	// 通常：ディレクトリ単位でランダム再生
					RandomList[NumRandomList].randomtype = 0;
					RandomList[NumRandomList].dirno = i;
					RandomList[NumRandomList].parentdirno = DirList[i].recno_parentdir;
					RandomList[NumRandomList].trackno = 0;
					NumRandomList++;
				}
				
				NumSndDir++;
			}
		}
	}
	
	return 0;
}


// DIRLIST配列の並べ替え
int compare_DIRLIST(const void *a, const void *b){
	DIRLIST *p1 = (DIRLIST *) a;
	DIRLIST *p2 = (DIRLIST *) b;
	
	if (p1 != NULL && p2 != NULL){
		if (p1->recno_parentdir < p2->recno_parentdir) return -1;
		else if (p1->recno_parentdir > p2->recno_parentdir) return 1;
		//else return strcmp(p1->fulldir, p2->fulldir);
		else return memcmp(p1->displayname, p2->displayname, DL_DISPLAYNAME_LEN + 1);
	}
	return 0;
}

/*
 *	DIR構成ファイルを書き出す
 */
int WriteDirStructureList(){
	FILE			*fout;
	char			rec[RECLEN_DIRLIST];
	char			dirname[DL_DIRNAME_LEN];
	int				i;
	
	printf("WriteDirectoryStructureList\n");
	// DIR構成ファイル出力
	// ０レコード目はヘッダレコード
	fout = fopen(DIRSTRUCTUREFILE14, "w");
	memset(rec, '\0', RECLEN_DIRLIST);
	fwrite(&DirListDetailRecNum, 2, 1, fout);		// ヘッダ
	fwrite(&NumSndDir, 2, 1, fout);
	fwrite(CrDateTime, 1, DL_CRDATETIME_LEN, fout);
	for (i=0; i<750; i++){
		fputc(0, fout);
	}
	
	// DIR名イメージファイル
	// １レコード目からがDIRレコード（その最初のレコードはルート）
	for (i=0; i<DirListDetailRecNum; i++){
		//_tprintf(_T("%-12s %3d %04d %03d %04d %04d %04d %02d\n"), 
		//	DirList[i].dirname, i, DirList[i].recno_subdir_start, DirList[i].num_subdir, 
		//	DirList[i].recno_parentdir, DirList[i].recno_grandparentdir, DirList[i].recno_nextrandomplay, DirList[i].num_sndfile);

		if (DirList[i].recno_subdir_start != 0){	// サブディレクトリがある
			DirList[i].recno_subdir_start += 1;		// ヘッダレコードの分を加算
		}
		if (DirList[i].num_subdir > 0x00FF){
			printf("サブDIR数制限超過 %s %d\n", DirList[i].fulldir, DirList[i].num_subdir);
			DirList[i].num_subdir = 0x00FF;	// 最初の２５５だけ有効に。
			// 制限超過していてもランダム再生の対象にはなる。
			// DIR選択機能で選択ができない。
		}
		DirList[i].recno_parentdir += 1;	// ヘッダレコードの分を加算（ルートディレクトリは自分自身）
		DirList[i].recno_grandparentdir += 1;	// ヘッダレコードの分を加算（ルートディレクトリは自分自身）
		
		memset(dirname, 0, DL_DIRNAME_LEN);
		memcpy(dirname, DirList[i].fulldir + DirList[i].idx_dirname, DirList[i].lenfulldir - DirList[i].idx_dirname);
		
		fwrite(&DirList[i].displayname, DL_DISPLAYNAME_LEN, 1, fout);
		fwrite(DirList[i].fulldir, DL_FULLDIR_LEN, 1, fout);
		fwrite(&DirList[i].lenfulldir, DL_LENFULLDIR_LEN, 1, fout);
		fwrite(dirname, DL_DIRNAME_LEN, 1, fout);
		fwrite(&DirList[i].recno_subdir_start, DL_RECNO_SUBDIR_LEN, 1, fout);
		fwrite(&DirList[i].num_subdir, DL_NUM_SUBDIR_LEN, 1, fout);
		fwrite(&DirList[i].recno_parentdir, DL_RECNO_PARENTDIR_LEN, 1, fout);
		fwrite(&DirList[i].recno_grandparentdir, DL_RECNO_GRANDPARENTDIR_LEN, 1, fout);
		fwrite(&DirList[i].num_sndfile, DL_NUM_SNDFILE_LEN, 1, fout);
		fwrite(&DirList[i].recno_sndfile, DL_RECNO_SNDFILE_LEN, 1, fout);
		fwrite(rec, DL_FILLER_14, 1, fout);
		
	}
	fclose(fout);

	return 0;
}


/*
 *	音源ファイル(MP3)リスト作成
 */
uint16_t MakeSndFileList(char *inFolder){
	DIR			*dirp;
	struct dirent *dp;
	char		*ext;
	FILE		*fout;
	char		outfile[DIRPATH_LEN + MP3FILE_LEN + 1];
	char		title[256];
	int			titlebytes;
	uint16_t	num = 0;
	MP3LIST14	rec04[0x100];
	char		rec[RECLEN_MP3LIST];
	int			i, j, k, len;
	char		TitleNamePrefix[ML_TITLENAME_LEN + 1];
	int			PrefixLen = 0;
	
	//printf("MakeSndFileList %s\n", inFolder);
	dirp = opendir(inFolder);
	if (dirp == NULL) return 0;
	
	// タグ情報取得テストの出力ファイル
	//sprintf(outfile, "%s/%s", inFolder, SNDTAGINFO);
	//fout = fopen(outfile, "w");
	
	memset(rec, 0, RECLEN_MP3LIST);
	
	while ((dp = readdir(dirp)) != NULL){
		//printf("MakeSndFileList=%s\n", dp->d_name);
		if (dp->d_type == DT_REG){
			ext = strrchr(dp->d_name, '.');		// 拡張子
			if (ext != NULL){
				if (strcmp(ext, ".mp3") == 0 || strcmp(ext, ".MP3") == 0){
					if (strlen(dp->d_name) > MP3FILE_LEN){
						continue;
					}
					if (num == 0x00FF){
						printf("MP3ファイル数制限超過 %s\n", inFolder);
						break;
					}
					
					// タグ情報取得テスト
					//	printf("getID3TagInfoTest start\n");
					//getID3TagInfoTest(fout, inFolder, dp->d_name);
					//	printf("getID3TagInfoTest end\n");
					
					memset(&rec04[num], 0, sizeof(MP3LIST14));
					strcpy(rec04[num].FileName, dp->d_name);
					memset(rec04[num].TitleName, ' ', ML_TITLENAME_LEN);
					rec04[num].TitleName[ML_TITLENAME_LEN] = '\0';
					
					num++;
				}
			}
		}
	}
	closedir(dirp);
	//fclose(fout);
	
	//printf("num=%d\n", num);
	if (num > 0){
		// ファイル名の順にソート
		qsort(rec04, num, sizeof(MP3LIST14), compare_MP3LIST14);
		// タグ情報からタイトル取得
		for (i=0; i<num; i++){
			sprintf(title, "%02d ", i + 1);
			len = strlen(title);
			getID3TagInfo(inFolder, rec04[i].FileName, title + len, 255 - len, &titlebytes);
			if (titlebytes > 0){
				title[titlebytes + len] = '\0';
				//printf("  UTF8 : %s\n", title);
				iconvUTF8toSJIS(title, titlebytes + len, rec04[i].TitleName, ML_TITLENAME_LEN);
			} else {
				ext = strrchr(rec04[i].FileName, '.');		// 拡張子
				iconvUTF8toSJIS(rec04[i].FileName, (int)(ext - rec04[i].FileName), rec04[i].TitleName, ML_TITLENAME_LEN);
			}
			// 全ファイル名の共通プレフィックス取得
			if (i == 1){
				for (j=0; j<ML_TITLENAME_LEN; j++){
					if (rec04[0].TitleName[j] == '\0' || rec04[1].TitleName[j] == '\0' ||
						rec04[0].TitleName[j] == '.' || rec04[1].TitleName[j] == '.'){
						break;
					} else if (rec04[0].TitleName[j] == rec04[1].TitleName[j]){
						TitleNamePrefix[j] = rec04[0].TitleName[j];
					} else {
						TitleNamePrefix[j] = '\0';
						PrefixLen = j;
						break;
					}
				}
			} else if (i > 1 && PrefixLen > 0){
				for (j=0; j<PrefixLen; j++){
					if (rec04[i].TitleName[j] == '\0' || rec04[i].TitleName[j] == '.'){
						PrefixLen = 0;
						break;
					} else if (rec04[i].TitleName[j] == TitleNamePrefix[j]){
					} else {
						TitleNamePrefix[j] = '\0';
						PrefixLen = j;
						break;
					}
				}
			}
		}
		// 共通プレフィックスがあるときは、それを除いた部分をタイトル名とする
		if (PrefixLen > 0){
			char	TitleNameNoPrefix[ML_TITLENAME_LEN + 1];	// 表示名
			for (i=0; i<num; i++){
				memset(TitleNameNoPrefix, ' ', ML_TITLENAME_LEN);
				TitleNameNoPrefix[ML_TITLENAME_LEN] = '\0';
				memcpy(TitleNameNoPrefix, rec04[i].TitleName + PrefixLen, ML_TITLENAME_LEN - PrefixLen);
				memcpy(rec04[i].TitleName, TitleNameNoPrefix, ML_TITLENAME_LEN);
			}
		}
	}
	
	// 音源ファイルリストファイル出力
	for (i=0; i<num; i++){
		fwrite(&rec04[i].TitleName, ML_TITLENAME_LEN, 1, Fml);
		fwrite(&rec04[i].FileName, ML_FILENAME_LEN, 1, Fml);
		fwrite(rec, ML_FILLER_14, 1, Fml);
		
	}
	
	return num;
}


// MP3LIST14配列の並べ替え
int compare_MP3LIST14(const void *a, const void *b){
	MP3LIST14 *p1 = (MP3LIST14 *) a;
	MP3LIST14 *p2 = (MP3LIST14 *) b;
	
	if (p1 != NULL && p2 != NULL){
		return strcmp(p1->FileName, p2->FileName);
	}
	return 0;
}


/*
 *	mp3ファイルのタグ情報から必要なものを取得
 *		title : 曲のタイトルをUTF8で取得する。末尾にNULL文字をセットしない。
 *		titlebufsize : titleバッファのサイズ
 *		*titlebytes : タイトルのバイト数。タイトルを取得できないときはマイナス。
 *		
 *	https://www.mars.org/mailman/public/mad-dev/2002-January/000439.html
 */
void getID3TagInfo(char *inFolder, char *mp3, char *title, int titlebufsize, int *titlebytes){
	char		mp3file[DIRPATH_LEN + MP3FILE_LEN + 1];
	struct id3_file *id3_file;
	struct id3_tag *id3_tag;
	struct id3_frame *id3_frame;
	int			i, j, k;
	
	*titlebytes = -1;
	
	sprintf(mp3file, "%s/%s", inFolder, mp3);
	id3_file = id3_file_open(mp3file, ID3_FILE_MODE_READONLY);
	id3_tag = id3_file_tag(id3_file);
	for (i=0; i<id3_tag->nframes; i++){
		id3_frame = id3_tag->frames[i];
		if (id3_frame != NULL){
			if (strcmp(id3_frame->id, ID3_FRAME_TITLE) == 0){
				for (j=0; j<id3_frame->nfields; j++){
					union id3_field		id3_field;
					id3_field = id3_frame->fields[j];
					
					if (id3_field.type == ID3_FIELD_TYPE_STRINGLIST){
						for (k=0; k<id3_field_getnstrings(&id3_field); k++){
							id3_ucs4_t *ucs4_str = id3_field_getstrings(&id3_field, k);
							*titlebytes = iconvUCS4toUTF8(ucs4_str, title, titlebufsize);
						}
					}
				}
			}
		}
	}
	id3_file_close(id3_file);
}

/*
 *	mp3ファイルのID3タグ情報を取得するテストプログラム
 *	(/usr/local/include/id3tag.h参照)
 *	https://www.mars.org/mailman/public/mad-dev/2002-January/000439.html
 */
void getID3TagInfoTest(FILE *fout, char *inFolder, char *mp3){
	char		mp3file[DIRPATH_LEN + MP3FILE_LEN + 1];
	struct id3_file *id3_file;
	struct id3_tag *id3_tag;
	struct id3_frame *id3_frame;
	char		utfbuf[512];
	char		sjisbuf[512];
	int			i, j, k;
	
	sprintf(mp3file, "%s/%s", inFolder, mp3);
	id3_file = id3_file_open(mp3file, ID3_FILE_MODE_READONLY);
	id3_tag = id3_file_tag(id3_file);
	//printf("id3_tag nframes=%d\n", id3_tag->nframes);
	for (i=0; i<id3_tag->nframes; i++){
		//signed long encoding = 0;
		
		id3_frame = id3_tag->frames[i];
		if (id3_frame != NULL){
			//printf("%d ID=%s desc=%s nfields=%d\n", i, id3_frame->id, id3_frame->description, id3_frame->nfields);
			fprintf(fout, "Frame Number = %d ID=%s\n", i, id3_frame->id);
			for (j=0; j<id3_frame->nfields; j++){
				union id3_field		id3_field;
				id3_field = id3_frame->fields[j];
				//printf(" field=%d id3_field_type=%d \n", j, id3_field.type);
				switch (id3_field.type){
				case ID3_FIELD_TYPE_TEXTENCODING: 
					//printf("ID3_FIELD_TYPE_TEXTENCODING "); 
					//printf("value=%ld ", id3_field.number.value);
					//encoding = id3_field.number.value;	// encodingを覚えておく必要はない
					break;
				case ID3_FIELD_TYPE_LATIN1: 
					//printf("ID3_FIELD_TYPE_LATIN1 "); 
					fprintf(fout, " ID3_FIELD_TYPE_LATIN1 : %s\n", id3_field.latin1.ptr);
					break;
				case ID3_FIELD_TYPE_LATIN1FULL: 
					//printf("ID3_FIELD_TYPE_LATIN1FULL "); 
					fprintf(fout, " ID3_FIELD_TYPE_LATIN1FULL : %s\n", id3_field.latin1.ptr);
					break;
				case ID3_FIELD_TYPE_LATIN1LIST: 
					//printf("ID3_FIELD_TYPE_LATIN1LIST "); 
					for (k=0; k<id3_field.latin1list.nstrings; k++){
						fprintf(fout, " ID3_FIELD_TYPE_LATIN1LIST : %s\n", id3_field.latin1list.strings[k]);
					}
					break;
				case ID3_FIELD_TYPE_STRING: 
					//printf("ID3_FIELD_TYPE_STRING "); 
					//fprintf(fout, " ID3_FIELD_TYPE_STRING : %s\n", id3_field.string.ptr);
					{
						id3_ucs4_t *ucs4_str = id3_field_getstring(&id3_field);
						//if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							//int bytes = convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							//iconvUTF16LEtoSJIS(utfbuf, bytes, sjisbuf, 128);
							int bytes = iconvUCS4toUTF8(ucs4_str, utfbuf, 512);
							bytes = iconvUTF8toSJIS(utfbuf, bytes, sjisbuf, 512);
							sjisbuf[bytes] = '\0';
							fprintf(fout, " ID3_FIELD_TYPE_STRING : %s\n", sjisbuf);
						//} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
						//	convUCS4toASCII(ucs4_str, sjisbuf, 128);
						//	fprintf(fout, " ID3_FIELD_TYPE_STRING : %s\n", sjisbuf);
						//}
					}
					break;
				case ID3_FIELD_TYPE_STRINGFULL: 
					//printf("ID3_FIELD_TYPE_STRINGFULL "); 
					//fprintf(fout, " ID3_FIELD_TYPE_STRINGFULL : %s\n", id3_field.string.ptr);
					{
						id3_ucs4_t *ucs4_str = id3_field_getfullstring(&id3_field);
						//if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							//int bytes = convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							//iconvUTF16LEtoSJIS(utfbuf, bytes, sjisbuf, 128);
							int bytes = iconvUCS4toUTF8(ucs4_str, utfbuf, 512);
							bytes = iconvUTF8toSJIS(utfbuf, bytes, sjisbuf, 512);
							sjisbuf[bytes] = '\0';
							fprintf(fout, " ID3_FIELD_TYPE_STRINGFULL : %s\n", sjisbuf);
						//} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
						//	convUCS4toASCII(ucs4_str, sjisbuf, 128);
						//	fprintf(fout, " ID3_FIELD_TYPE_STRINGFULL : %s\n", sjisbuf);
						//}
					}
					break;
				case ID3_FIELD_TYPE_STRINGLIST: 
					//printf("ID3_FIELD_TYPE_STRINGLIST "); 
					//printf("nstrings=%d ", id3_field.stringlist.nstrings);
					for (k=0; k<id3_field_getnstrings(&id3_field) /*id3_field.stringlist.nstrings*/; k++){
						id3_ucs4_t *ucs4_str = id3_field_getstrings(&id3_field, k);
						//if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							//int bytes = convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							//iconvUTF16LEtoSJIS(utfbuf, bytes, sjisbuf, 128);
							int bytes = iconvUCS4toUTF8(ucs4_str, utfbuf, 512);
							bytes = iconvUTF8toSJIS(utfbuf, bytes, sjisbuf, 512);
							sjisbuf[bytes] = '\0';
							fprintf(fout, " ID3_FIELD_TYPE_STRINGLIST : %s\n", sjisbuf);
						//} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
						//	printf("  convUCS4toASCII start\n");
						//	convUCS4toASCII(ucs4_str, sjisbuf, 128);
						//	printf("  convUCS4toASCII end\n");
						//	fprintf(fout, " ID3_FIELD_TYPE_STRINGLIST : %s\n", sjisbuf);
						//}
					}
					break;
				case ID3_FIELD_TYPE_LANGUAGE: 
					//printf("ID3_FIELD_TYPE_LANGUAGE "); 
					/*
					{
						id3_ucs4_t *ucs4_str = id3_field.string.ptr;
						if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							iconvUTF16LEtoSJIS(utfbuf, 128, sjisbuf, 128);
							fprintf(fout, " ID3_FIELD_TYPE_LANGUAGE : %s\n", sjisbuf);
						} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
							convUCS4toASCII(ucs4_str, sjisbuf, 128);
							fprintf(fout, " ID3_FIELD_TYPE_LANGUAGE : %s\n", sjisbuf);
						}
					}
					*/
					break;
				case ID3_FIELD_TYPE_FRAMEID: 
					//printf("ID3_FIELD_TYPE_FRAMEID "); 
					break;
				case ID3_FIELD_TYPE_DATE: 
					//printf("ID3_FIELD_TYPE_DATE "); 
					break;
				case ID3_FIELD_TYPE_INT8: 
					//printf("ID3_FIELD_TYPE_INT8 "); 
					break;
				case ID3_FIELD_TYPE_INT16: 
					//printf("ID3_FIELD_TYPE_INT16 "); 
					break;
				case ID3_FIELD_TYPE_INT24: 
					//printf("ID3_FIELD_TYPE_INT24 "); 
					break;
				case ID3_FIELD_TYPE_INT32: 
					//printf("ID3_FIELD_TYPE_INT32 "); 
					break;
				case ID3_FIELD_TYPE_INT32PLUS: 
					//printf("ID3_FIELD_TYPE_INT32PLUS "); 
					break;
				case ID3_FIELD_TYPE_BINARYDATA: 
					//printf("ID3_FIELD_TYPE_BINARYDATA "); 
					break;
				}
				//printf(" \n");
			}
		}
	}
	id3_file_close(id3_file);
}


/*
 *	ランダム再生のオーダーを決める
seed=1488681854
seed=1488681891
seed=1488728475
 */
int SetRandomOrder(){
	int				shuffle_cnt;
	int				i, j;
	unsigned int	seed;
	int 			idx1, idx2;
	RANDOMLIST		d;

	printf("random order start\n");
	// ランダム再生順決定
	shuffle_cnt = (int)NumRandomList*1000;
	seed = (unsigned)time(NULL);
	printf("seed=%d\n", seed);
	srand(seed);	// seedはこれでよいか
	// 
	for (i=0; i<shuffle_cnt; i++){
		idx1 = (rand() % NumRandomList);
		idx2 = (rand() % NumRandomList);
		if (idx1 == idx2) continue;
		
		d = RandomList[idx1];
		RandomList[idx1] = RandomList[idx2];
		RandomList[idx2] = d;
	}
	
	for (i=1; i<(NumRandomList - 1); i++){
		if (RandomList[i].randomtype == RANDOMTYPE_FILE && RandomList[i].dirno == RandomList[i - 1].dirno){
			// トラック単位のランダム再生では同じディレクトリの曲が連続することを避ける
			// うしろからディレクトリを異にする要素を探す
			for (j=i+1; j<NumRandomList; j++){
				if (RandomList[i].dirno != RandomList[j].dirno){
					// 入れ替え
					d = RandomList[i];
					RandomList[i] = RandomList[j];
					RandomList[j] = d;
					break;
				}
			}
		} else if (RandomList[i].randomtype == RANDOMTYPE_FOLDER && RandomList[i].parentdirno == RandomList[i - 1].parentdirno){
			// フォルダ単位のランダム再生では親ディレクトリの連続を避ける
			// うしろから親ディレクトリを異にする要素を探す
			for (j=i+1; j<NumRandomList; j++){
				if (RandomList[i].parentdirno != RandomList[j].parentdirno){
					// 入れ替え
					d = RandomList[i];
					RandomList[i] = RandomList[j];
					RandomList[j] = d;
					break;
				}
			}
		}
	}
	
	return 0;
}


/*
 *	ランダム再生順ファイルを出力
 */
void WriteRandomOrderFile(){
	FILE			*fout;
	int				i;
	uint16_t		dirno;
	
	fout = fopen(RANDOMORDERFILE, "w");
	
	for (i=0; i<NumRandomList; i++){
		printf("%d : %d %d %s %d\n", i, RandomList[i].dirno, RandomList[i].randomtype, DirList[RandomList[i].dirno].fulldir, RandomList[i].trackno);
		fwrite(&RandomList[i].randomtype, RO_RANDOMTYPE_LEN, 1, fout);
		dirno = RandomList[i].dirno + 1;		// +1はディレクトリ構成ファイルのヘッダレコードの分
		fwrite(&dirno, RO_DIRNO_LEN, 1, fout);
		fwrite(&RandomList[i].trackno, RO_TRACKNO_LEN, 1, fout);
	}
	
	fclose(fout);
}


/*
 *	UCS4文字列をUTF8文字列に変換。
 *		ucs4 : UCS4文字列。終端NULLを前提。
 *		utf8str : 変換されたUTF8文字列。終端にNULLをセットしない。
 *		戻り値 : utf8strにセットされた文字列のバイト数。エラー時はマイナス。
 */
int iconvUCS4toUTF8(id3_ucs4_t *ucs4, char *utf8str, int bufsize){
	char *p_src, *p_dst;
	size_t n_src, n_dst;
	int	rc;
	int bytes = 0;
	
	p_dst = utf8str;
	n_dst = bufsize;
	
	while (*ucs4 != 0){
		p_src = (char *)ucs4;
		n_src = 4;
		if ((rc = iconv(icdUCS4toUTF8, &p_src, &n_src, &p_dst, &n_dst)) == -1){
			return -1;
		}
		bytes += rc;
		ucs4++;
	}
	
	return (bufsize - n_dst);
}

/*
 *	UTF8文字列をSJIS文字列に変換
 *		utf8str : 変換元文字列
 *		convbytes : 変換対象バイト数
 *		sjisstr : 返還後文字列用変数（末尾にNULLを付けない）
 *		sjismaxlen : 最大変換後バイト数
 */
int iconvUTF8toSJIS(char *utf8str, int convbytes, char *sjisstr, int sjismaxlen){
	char *p_src, *p_dst;
	size_t n_src, n_dst;
	int	rc;
	
	p_src = utf8str;
	p_dst = sjisstr;
	n_src = convbytes;
    n_dst = sjismaxlen;
	while(0 < n_src && 0 < n_dst){
		if ((rc = iconv(icdUTF8toSJIS, &p_src, &n_src, &p_dst, &n_dst)) == -1){
			return -1;
		}
	}
	return (sjismaxlen - n_dst);
}


/*
 *	UTF16LE文字列をSJIS文字列に変換
 *		utf8str : 変換元文字列
 *		convbytes : 変換対象バイト数
 *		sjisstr : 返還後文字列用変数
 *		sjismaxlen : 最大変換後バイト数
 */
int iconvUTF16LEtoSJIS(char *utf16str, int convbytes, char *sjisstr, int sjismaxlen){
	char *p_src, *p_dst;
	size_t n_src, n_dst;
	int	rc;
	
	p_src = utf16str;
	p_dst = sjisstr;
	n_src = convbytes;
    n_dst = sjismaxlen;
	while(0 < n_src && 0 < n_dst){
		//printf("   %ld %ld\n", n_src, n_dst);
		if ((rc = iconv(icdUTF16LEtoSJIS, &p_src, &n_src, &p_dst, &n_dst)) == -1){
			break;
		}
	}
	*p_dst = '\0';
	
	return rc;
}


// ucs4文字列をUTF16LE文字列に変換
// UTF16LE文字列のバイト数を返す
int convUCS4toUTF16LE(id3_ucs4_t *ucs4, char *utf16lestr, int bufsize){
	int bytes = 0;
	
	while (*ucs4 != 0 && bufsize > 2){
		*utf16lestr++ = (*ucs4 & 0x00FF);
		bufsize--;
		*utf16lestr++ = ((*ucs4 >> 8) & 0x00FF);
		bufsize--;
		ucs4++;
		bytes += 2;
	}
	*utf16lestr++ = 0;
	*utf16lestr++ = 0;
	
	return bytes;
}

// ucs4文字列をASCII文字列に変換
void convUCS4toASCII(id3_ucs4_t *ucs4, char *asciistr, int bufsize){
	while (*ucs4 != 0 && bufsize > 1){
		*asciistr++ = (*ucs4 & 0x00FF);
		bufsize--;
		ucs4++;
	}
	*asciistr++ = 0;
}


// 履歴ファイル新規作成
void CreateHistFile(){
	FILE		*fout;
	int			i;
	uint16_t	lastdir = 0;
	uint8_t		lasttrack = 0;
	uint32_t	lastbyte = 0;
	
	fout = fopen(HISTFILE, "w");
	for (i=0; i<RECLEN_HISTFILE; i++){
		fputc(0, fout);
	}
	fclose(fout);
}


/*
 *	環境変数ENV_RANDOMを取得
 */
char *GetEnvRandomDir(){
	int		i;
	char	**e;
	int		len = strlen(ENV_RANDOM);
	
	for (i=0, e=environ; *e != NULL; i++, e++){
		if (strncmp(*e, ENV_RANDOM, len) == 0 && *(*e + len) == '='){
			printf("env=[%s]\n", *e + len + 1);
			return *e + len + 1;
		}
	}
	
	return NULL;
}

