﻿/*
 *	Source file	: mp3_14_serv.c
 *	Date		: 2019/10/13
 *	Update		: 2019/12/01
 *	mp3_14_menuとWindows版コントローラとの間の通信を中継。
 *	
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <poll.h>
#include <signal.h>

#define	PORT_TO_CLIENT					"5000"
#define	ENV_PORT_TO_CTRL				"PORT_TO_CTRL"		// ポート番号を環境変数から取得する (mp3_14.sh)
#define	WINCTRL_CMDLENGTH				16					// コントローラとの通信１単位のバイト数
#define POLL_TIMEOUT					50					// poll関数のタイムアウト (ms)

void InterruptSIGPIPE(int signaltype);
int writeall(int fd, void *buf, size_t nbyte);
int readall(int fd, void *buf, size_t nbyte);
char *GetEnvPortToCtrl();

int		fdWin;			// ウインドウズ版コントローラとの通信

extern char **environ;		// 環境変数取得

int main(int argc, char *argv[]){
	struct addrinfo	hint;
	struct addrinfo	*res;
	struct sockaddr_storage	sa;
	int		sock;
	int		error;
	socklen_t	salen;
	char	serv[NI_MAXSERV];
	fd_set	ready;
	char	buf[WINCTRL_CMDLENGTH];
	ssize_t	rval;
	int		i;
	int		fdmenu;			// mp3_14_menuとの通信
	struct pollfd fd_poll[1];
	struct sigaction	handler;
	struct timeval	selecttimeout;
	int		selcnt;
	char	*p_port;
	char	port_to_ctrl[16];		// コントローラとの通信ポート
	
	handler.sa_handler = InterruptSIGPIPE;
	if (sigfillset(&handler.sa_mask) < 0){
		printf("sigfillset error\n");
	}
	handler.sa_flags = 0;
	if (sigaction(SIGPIPE, &handler, 0) < 0){
		printf("sigaction error\n");
	}
	
	// mp3_14_menuから渡されるパラメータ
	for (i=0; i<argc; i++){
		printf("mp3_14_serv param %d : %s\n", i, argv[i]);
	}
	if (argc > 1){
		fdmenu = atoi(argv[1]);		// socketpairの片方
	} else {
		fdmenu = 0;
	}
	// TCPポート番号
	p_port = GetEnvPortToCtrl();
	if (p_port != NULL){
		strcpy(port_to_ctrl, p_port);
	} else {
		strcpy(port_to_ctrl, PORT_TO_CLIENT);
	}
	
	// ウインドウズ版コントローラとの通信のためのsocket
	memset(&hint, 0, sizeof(hint));
	hint.ai_family = PF_INET;
	hint.ai_socktype = SOCK_STREAM;
	hint.ai_flags = AI_PASSIVE;
	error = getaddrinfo(NULL, port_to_ctrl /*PORT_TO_CLIENT*/, &hint, &res);
	if (error){
		printf("%s\n", gai_strerror(error));
		return 0;
	}
	if (res->ai_addr == NULL){
		printf("cannot get address\n");
	}
	if ((sock = socket(res->ai_addr->sa_family, SOCK_STREAM, 0)) < 0){		// ★
		printf("socket error\n");
	}
	if (bind(sock, res->ai_addr, res->ai_addrlen)){			// ★
		printf("bind error\n");
	}
	
	salen = sizeof(sa);
	if (getsockname(sock, (struct sockaddr *)&sa, &salen)){		// ★
		printf("getsockname error\n");
	}
	getnameinfo((struct sockaddr *)&sa, salen, NULL, 0, serv, NI_MAXSERV, NI_NUMERICSERV);		// ★
	printf("port number=%s\n", serv);
	
	printf("listen start\n");
	listen(sock, 1);				// ★
	printf("listen end\n");
	fdWin = -1;
	
	for (;;){
		FD_ZERO(&ready);
		FD_SET(sock, &ready);
		if (fdWin >= 0){
			FD_SET(fdWin, &ready);
		}
//		printf("select start\n");
		selecttimeout.tv_sec = 0;
		selecttimeout.tv_usec = 500000;
		if ((selcnt=select(32, &ready, 0, 0, &selecttimeout)) < 0){		// ★
			printf("select error\n");
			continue;
		}
//		printf("select end\n");
		if (selcnt > 0 && FD_ISSET(sock, &ready)){
			if (fdWin == -1){
				printf("connect 1 \n");
				if ((fdWin = accept(sock, 0, 0)) == -1){		// ★
					printf("accept 1 error\n");
					break;
				}
				printf("fdWin=%d \n", fdWin);
			} else {
				int fd_tmp;
				printf("cannot connect 2nd\n");
				if ((fd_tmp = accept(sock, 0, 0)) != -1){		// ★
					close(fd_tmp);
				}
			}
		}
		// ウインドウズ版コントローラからの受信をmp3_14_menuへ送信
		if (fdWin >= 0 && FD_ISSET(fdWin, &ready)){
			if ((rval = read(fdWin, buf, WINCTRL_CMDLENGTH)) > 0){		// ★
				//printf("read 1 :%d\n", rval);
				if (fdmenu != 0){
					writeall(fdmenu, buf, rval);
				}
			} else if (rval == 0){
				printf("fdWin rval==0\n");
				close(fdWin);
				fdWin = -1;
			}
		}
		// mp3_14_menuからの受信をウインドウズ版コントローラへ送信
		//printf("poll start\n");
		memset(&fd_poll, 0, sizeof(fd_poll));
		fd_poll[0].fd = fdmenu;
		fd_poll[0].events = POLLIN | POLLERR;
		poll(fd_poll, 1, POLL_TIMEOUT);
		//printf("poll end\n");
		if (fd_poll[0].revents & POLLIN) {
			if ((rval = readall(fdmenu, buf, WINCTRL_CMDLENGTH)) > 0){
				printf(" mp3_14_serv : from menu %x\n", buf[0]);
				if (fdWin != -1){
					rval = write(fdWin, buf, rval);
					//printf("write rval==%d\n", rval);
					if (rval <= 0){
						close(fdWin);
						fdWin = -1;
					}
				}
			}
//		} else {
//			printf(" mp3_14_serv : read from mp3_14_host NO DATA\n");
		}
	}
}


void InterruptSIGPIPE(int signaltype){
	printf("InterruptSIGPIPE signaltype==%d\n", signaltype);
	close(fdWin);
	fdWin = -1;
	return;
}


/*
 *	所定バイトをすべてwriteするまで繰り返す
 */
int writeall(int fd, void *buf, size_t nbyte){
	ssize_t	wsz = 0;
	int		total = 0;
	
	do {
		wsz = write(fd, buf + total, nbyte - total);
		if (wsz >= 0){
			total += wsz;
		} else {
			return total;
		}
	} while (total < nbyte);
	
	return total;
}


/*
 *	所定バイトをすべてreadするまで繰り返す
 */
int readall(int fd, void *buf, size_t nbyte){
	ssize_t	wsz = 0;
	int		total = 0;
	
	do {
		wsz = read(fd, buf + total, nbyte - total);
		if (wsz >= 0){
			total += wsz;
		} else {
			return total;
		}
	} while (total < nbyte);
	
	return total;
}


/*
 *	環境変数ENV_PORT_TO_CTRLを取得
 */
char *GetEnvPortToCtrl(){
	int		i;
	char	**e;
	int		len = strlen(ENV_PORT_TO_CTRL);
	
	for (i=0, e=environ; *e != NULL; i++, e++){
		if (strncmp(*e, ENV_PORT_TO_CTRL, len) == 0 && *(*e + len) == '='){
			printf("env=[%s]\n", *e + len + 1);
			return *e + len + 1;
		}
	}
	
	return NULL;
}

